# Greek Alphabet

greek = [r'\alpha',  r'\beta', r'\gamma', r'\delta',r'\mu', r'\nu',r'\varrho', r'\rho',r'\sigma',r'\tau', r'\omega', r'\kappa', r'\lambda',r'\psi', r'\eta',r'\vartheta',  r'\theta', r'\iota',r'\zeta', r'\omicron', r'\pi', r'\upsilon', r'\varphi',r'\phi',r'\chi']

greek_dict = {'alpha' : r'\alpha', 'beta' : r'\beta', 'gamma' : r'\gamma', 'delta': r'\delta', 'mu' : r'\mu', 'nu' : r'\nu','varrho' : r'\rho', 'rho' : r'\rho', 'sigma' : r'\sigma', 'tau' : r'\tau', 'omega' : r'\omega', 'kappa' : r'\kappa', 'lambda' : r'\lambda','psi': r'\psi', 'eta' : r'\eta','vartheta' : r'\theta', 'theta' : r'\theta', 'iota': r'\iota', 'zeta' : r'\zeta', 'omicron' : r'\omicron', 'pi' : r'\pi', 'upsilon' : r'\upsilon','varphi' : r'\phi', 'phi': r'\phi', 'chi': r'\chi',\
    'Alpha' : r'\Alpha', 'Beta' : r'\Beta', 'Gamma' : r'\Gamma', 'Delta': r'\Delta', 'Mu' : r'\Mu', 'Nu' : r'\Nu', 'Rho' : r'\Rho', 'Sigma' : r'\Sigma', 'Tau' : r'\Tau', 'Omega' : r'\Omega', 'Kappa' : r'\Kappa', 'Lambda' : r'\Lambda', 'Eta' : r'\Eta', 'Theta' : r'\Theta', 'Iota': r'\Iota', 'Zeta' : r'\Zeta', 'Omicron' : r'\Omicron', 'Pi' : r'\Pi', 'Upsilon' : r'\Upsilon', 'Phi' : r'\Phi', 'Chi' : r'\Chi', 'Psi' : r'\Psi'}
